package artificialLifeSimulator.GUI;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ConcurrentLinkedQueue;

import artificialLifeSimulator.GUI.controls.ZoomEventHandler;
import artificialLifeSimulator.lifeForms.populations.BasicPlantPopulation;
import artificialLifeSimulator.lifeForms.populations.Biome;
import javafx.animation.AnimationTimer;
import javafx.application.Platform;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.scene.canvas.Canvas;
import javafx.scene.canvas.GraphicsContext;
import javafx.scene.control.Alert;
import javafx.scene.control.Alert.AlertType;
import javafx.scene.input.MouseButton;
import javafx.scene.input.MouseEvent;
import javafx.scene.layout.Pane;
import javafx.scene.paint.Color;

/**
 * @author Toni
 * Class is used to implement the pane where the biome is being visualized
 */
public class SimulationPane extends Pane implements Runnable{
	private BlockingQueue<Biome> drawingQueue;
	
	private double canvasWidth;
	private double canvasHeight;
	private double widthWindow;
	private double heightWindow;

	private double startingScaleX;
	private double startingScaleY;
	private double startingTranslationX;
	private double startingTranslationY;
	
	private Canvas topLayer;
	private Canvas midLayer;
	private Canvas bottomLayer;
	private GraphicsContext topLayer_g;
	private GraphicsContext midLayer_g;
	private GraphicsContext bottomLayer_g;
	
	private volatile Biome current;
	private volatile Thread exit;
	private volatile boolean block = false;
	
	/**
	 * Constructor method used to initialize the simulation pine
	 * @param canvasWidth width of the canvas
	 * @param canvasHeight height of the canvas
	 * @param widthWindow width of the window 
	 * @param heightWindow height of the window
	 * @param drawingQueue drawing queue used to communicate between drawing and simulating thread
	 */
	public SimulationPane(double canvasWidth, double canvasHeight, double widthWindow, double heightWindow, BlockingQueue<Biome> drawingQueue) {
		this.drawingQueue=drawingQueue;
		this.widthWindow = widthWindow;
		this.heightWindow = heightWindow;
		this.canvasWidth = canvasWidth;
		this.canvasHeight = canvasHeight;
		calculateStartingTransformations();
		generateLayers();
	}

	/**
	 *  Calculates initial transformations of the canvas
	 */
	private void calculateStartingTransformations() {
		this.startingScaleX = (1.0* this.widthWindow)/this.canvasWidth;
		this.startingScaleY = (1.0* this.heightWindow)/this.canvasHeight;
		this.startingTranslationX = ((canvasWidth - canvasWidth/startingScaleX)*(startingScaleX/(2)));
		this.startingTranslationY = ((canvasHeight - canvasHeight/startingScaleY)*(startingScaleY/(2)));
		
	}

	/**
	 *  Generates the drawing layers
	 *  top layer - moving organisms, herbivores and carnivores
	 *  mid layer - plants
	 *  bottom layer - herbivores
	 */
	private void generateLayers() {
		topLayer = new Canvas(canvasWidth, canvasHeight);
		midLayer = new Canvas(canvasWidth, canvasHeight);
		bottomLayer = new Canvas(canvasWidth, canvasHeight);
		topLayer_g = topLayer.getGraphicsContext2D();
		midLayer_g = midLayer.getGraphicsContext2D();
		bottomLayer_g = bottomLayer.getGraphicsContext2D();
		this.getChildren().removeAll();
		this.getChildren().add(topLayer);
		this.getChildren().add(midLayer);
		this.getChildren().add(bottomLayer);
		midLayer.toFront();
		topLayer.toFront();
		topLayer.addEventHandler(MouseEvent.MOUSE_PRESSED, new ZoomEventHandler(topLayer, midLayer, bottomLayer, canvasHeight, canvasWidth, widthWindow, heightWindow));
		
    	
    	topLayer.setScaleY(startingScaleY);
    	topLayer.setScaleX(startingScaleX);
    	topLayer.setTranslateX(startingTranslationX);
    	topLayer.setTranslateY(startingTranslationY);
    	

    	midLayer.setScaleY(startingScaleY);
    	midLayer.setScaleX(startingScaleX);
    	midLayer.setTranslateX(startingTranslationX);
    	midLayer.setTranslateY(startingTranslationY);

    	bottomLayer.setScaleY(startingScaleY);
    	bottomLayer.setScaleX(startingScaleX);
    	bottomLayer.setTranslateX(startingTranslationX);
    	bottomLayer.setTranslateY(startingTranslationY);
	}
	
	
	/**
	 * Method is used to kill the simulation thread 
	 */
	public void stop(){
		exit = null;
	}
	
	/**
	 * Method is used to pause the simulation thread
	 */
	public void block(){
		block = true;
	}
	
	/**
	 * Method is used to get the current Biome
	 * @return the current biome
	 */
	public Biome getCurrentBiome(){
		return current;
	}

	@Override
	public void run() {
		exit = Thread.currentThread();
		Thread firstThread = Thread.currentThread();
		int temp=0;
		while(exit == firstThread){
			try {
				Thread.sleep(GuiConstants.simulationPaneIterationSleep); 
			} catch (InterruptedException e) {

			}
			if(block){
				try {
					Thread.sleep(Long.MAX_VALUE);
				} catch (InterruptedException e) {
					block = false;
				}
			}
			

			try {

				current = drawingQueue.take();
				Platform.runLater(new Runnable(){

					@Override
					public void run() {
						topLayer_g.clearRect(0, 0, canvasWidth, canvasWidth);
						midLayer_g.clearRect(0, 0, canvasWidth, canvasWidth);
						bottomLayer_g.clearRect(0, 0, canvasWidth, canvasWidth);
						current.drawBiome(topLayer_g, midLayer_g, bottomLayer_g);
					}
					
				});
			} catch (InterruptedException e) {

			}
		}
	}
}
