package artificialLifeSimulator.GUI.controls;


import artificialLifeSimulator.GUI.GuiConstants;
import javafx.application.Platform;
import javafx.event.Event;
import javafx.event.EventHandler;
import javafx.scene.canvas.Canvas;
import javafx.scene.canvas.GraphicsContext;
import javafx.scene.input.MouseButton;
import javafx.scene.input.MouseEvent;
import javafx.scene.paint.Color;

/**
 * @author Toni
 * Class implements mouse event handler required for zoom
 */
public class ZoomEventHandler implements EventHandler<MouseEvent>{
	private Canvas topLayer;
	private Canvas midLayer;
	private Canvas bottomLayer;
	private GraphicsContext topLayer_g;
	private GraphicsContext middleLayer_g;
	private GraphicsContext bottomLayer_g;
	private double canvasHeight;
	private double canvasWidth;
	private double widthWindow;
	private double heightWindow;

	private double startingScaleX;
	private double startingScaleY;
	private double startingTranslationX;
	private double startingTranslationY;
	
	/**
	 * Constructor for zoom event handler
	 * @param topLayer top layer canvas where carnivores and herbivores are drawn
	 * @param middleLayer middle layer canvas where plants are drawn
	 * @param bottomLayer bottom layer canvas where habitat is drawn
	 * @param canvasHeight height of canvas
	 * @param canvasWidth width of canvas
	 * @param windowWidth window width where the canvases are set
	 * @param windowHeight window height where canvases are set
	 */
	public ZoomEventHandler(Canvas topLayer, Canvas middleLayer, Canvas bottomLayer, double canvasHeight, double canvasWidth, 
			double windowWidth, double windowHeight){
		this.topLayer = topLayer;
		this.topLayer_g = topLayer.getGraphicsContext2D();
		this.midLayer = middleLayer;
		this.middleLayer_g = middleLayer.getGraphicsContext2D();
		this.bottomLayer = bottomLayer;
		this.bottomLayer_g = bottomLayer.getGraphicsContext2D();
		this.canvasHeight = canvasHeight;
		this.canvasWidth = canvasWidth;
		this.heightWindow = windowHeight;
		this.widthWindow = windowWidth;
		calculateStartingTransformations();
	}

	/**
	 *  Calculates initial transformations of the canvas
	 */
	private void calculateStartingTransformations() {
		this.startingScaleX = (1.0* this.widthWindow)/this.canvasWidth;
		this.startingScaleY = (1.0* this.heightWindow)/this.canvasHeight;
		this.startingTranslationX = ((canvasWidth - canvasWidth/startingScaleX)*(startingScaleX/(2)));
		this.startingTranslationY = ((canvasHeight - canvasHeight/startingScaleY)*(startingScaleY/(2)));
		
	}
	
	@Override
	public void handle(MouseEvent event) {
		Platform.runLater(new Runnable(){

			@Override
			public void run() {
				if(MouseButton.PRIMARY == event.getButton()){
			    	topLayer.setScaleX(topLayer.getScaleX() + GuiConstants.GUIScalingFactor);
			    	topLayer.setScaleY(topLayer.getScaleY() + GuiConstants.GUIScalingFactor);
			    	
			    	midLayer.setScaleX(midLayer.getScaleX() + GuiConstants.GUIScalingFactor);
			    	midLayer.setScaleY(midLayer.getScaleY() + GuiConstants.GUIScalingFactor);

			    	bottomLayer.setScaleX(bottomLayer.getScaleX() + GuiConstants.GUIScalingFactor);
			    	bottomLayer.setScaleY(bottomLayer.getScaleY() + GuiConstants.GUIScalingFactor);
			    	
			    	double XTranslation = determineTranslation(topLayer.getScaleX(), event.getX(), canvasWidth, startingScaleX, widthWindow);
			    	double YTranslation = determineTranslation(topLayer.getScaleY(), event.getY(), canvasHeight, startingScaleY, heightWindow);
			    	
			    	topLayer.setTranslateY(YTranslation);
			    	topLayer.setTranslateX(XTranslation);
			    	
			    	midLayer.setTranslateY(YTranslation);
			    	midLayer.setTranslateX(XTranslation);
			    	
			    	bottomLayer.setTranslateY(YTranslation);
			    	bottomLayer.setTranslateX(XTranslation);
			    	
				}else {
					System.out.println(startingScaleY);
					System.out.println(startingScaleX);
					topLayer.setScaleY(startingScaleY);
			    	topLayer.setScaleX(startingScaleX);
			    	topLayer.setTranslateX(startingTranslationX);
			    	topLayer.setTranslateY(startingTranslationY);
			    	
			    	midLayer.setScaleY(startingScaleY);
			    	midLayer.setScaleX(startingScaleX);
			    	midLayer.setTranslateX(startingTranslationX);
			    	midLayer.setTranslateY(startingTranslationY);

			    	bottomLayer.setScaleY(startingScaleY);
			    	bottomLayer.setScaleX(startingScaleX);
			    	bottomLayer.setTranslateX(startingTranslationX);
			    	bottomLayer.setTranslateY(startingTranslationY);
			    	bottomLayer_g.fillRect(0, 0, canvasWidth, canvasHeight);
				}
			}
			
		});
	}
	
	/**
	 * Method is used to determine the translation for every canvas
	 * @param scaleFactor scaling factor for the axis
	 * @param mouseCoordinate mouse coordinate x or y
	 * @param canvasDim dimension of canvas width or height
	 * @param currentScale of zoom for the canvas axis
	 * @param windowDim axis dimension
	 * @return the new translation value for the axis
	 */
	public double determineTranslation(double scaleFactor, double mouseCoordinate, double canvasDim, double currentScale, double windowDim){
		System.out.println(canvasDim + " " + mouseCoordinate + " " + scaleFactor+ " " + windowDim);
		double trans=0;
		double centerTrans;
		centerTrans = ((canvasDim - canvasDim/currentScale)*(currentScale/(2)));

		if(mouseCoordinate>canvasDim/2){
			trans = -(mouseCoordinate-canvasDim/2)*scaleFactor+centerTrans;
			if(trans<-canvasDim/2*scaleFactor+windowDim/2+centerTrans){
				trans=-canvasDim/2*scaleFactor+windowDim/2+centerTrans;
			}
		}else{
			trans = (canvasDim/2 - mouseCoordinate)*scaleFactor+centerTrans;
			if(trans>canvasDim/2*scaleFactor-windowDim/2+centerTrans){
				trans=canvasDim/2*scaleFactor-windowDim/2+centerTrans;
			}
		}
		return trans;
	}

}
