package artificialLifeSimulator.lifeForms.basic;

import java.util.LinkedList;
import java.util.Random;

import artificialLifeSimulator.abstractClasses.AbstractHerbivore;
import artificialLifeSimulator.abstractClasses.AbstractPlant;
import artificialLifeSimulator.helperClasses.Location;
import javafx.scene.canvas.GraphicsContext;
import javafx.scene.paint.Color;

/**
 * @author Toni
 * Class implements methods that allow the simulation of a basic herbivore
 */
public class BasicHerbivore extends AbstractHerbivore<BasicHerbivore>{
	private double maxHerbivoreEnergy;
	private double currentEnergy;
	private double chanceOfReproduction;
	private double agePerIteration;
	private double maxAge;
	private double currentAge;
	private double startingEnergyOffsprings;
	private double energyForReproduction;
	private double maxSpeed;
	

	/**
	 * Constructor for Basic herbivore
	 * @param maxSpeed maximum herbivore speed
	 * @param maxEnergy maximum herbivore energy
	 * @param startingEnergy starting herbivore energy
	 * @param chanceOfReproduction chance of reproduction
	 * @param agePerIteration ages gained per iteration
	 * @param maxAge maximal age
	 * @param energyForReproduction energy required for reproduction
	 * @param location starting location
	 */
	public BasicHerbivore(double maxSpeed,double maxEnergy, double startingEnergy, double chanceOfReproduction, double agePerIteration,double maxAge,double energyForReproduction,
			Location location) {
		this.maxSpeed = maxSpeed;
		this.maxHerbivoreEnergy=maxEnergy;
		this.currentEnergy=startingEnergy;
		this.chanceOfReproduction=chanceOfReproduction;
		this.startingEnergyOffsprings=startingEnergy;
		this.agePerIteration=agePerIteration;
		this.currentLocation = location;
		this.currentAge=0;
		this.maxAge = maxAge;
		this.energyForReproduction=energyForReproduction;
	}
	
	/**
	 * Default constructor
	 */
	private BasicHerbivore(){
		
	}

	@Override
	public AbstractHerbivore newLikeThis() {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public AbstractHerbivore copyOf() {
		BasicHerbivore temp = new BasicHerbivore();
		temp.maxHerbivoreEnergy=this.maxHerbivoreEnergy;
		temp.currentEnergy=this.currentEnergy;
		temp.chanceOfReproduction=this.chanceOfReproduction;
		temp.agePerIteration=this.agePerIteration;
		temp.maxAge=this.maxAge;
		temp.currentAge=this.currentAge;
		temp.startingEnergyOffsprings=this.startingEnergyOffsprings;
		temp.energyForReproduction = this.energyForReproduction;
		temp.maxSpeed=this.maxSpeed;
		temp.currentLocation=this.currentLocation.copyOf();
		return temp;
	}

	@Override
	public void drawMe(GraphicsContext g) {
		double color =Math.abs( currentEnergy/maxHerbivoreEnergy);
		color = (1-color)==0 ? 0.2:color;
		g.setFill(new Color(0, 0, color, 1));
		g.fillOval(this.currentLocation.getX(), this.currentLocation.getY(), 10, 10);
		
	}

	@Override
	public LinkedList<BasicHerbivore> nextIteration(int ekosystemWidth, int ekosystemHeight, Random rand,
			AbstractPlant<AbstractPlant> closestPlant) {
		this.currentAge+=this.agePerIteration;
		this.currentEnergy-=0.05;
		if(this.currentAge>this.maxAge) return null;
		if(this.currentEnergy>(this.energyForReproduction/this.maxHerbivoreEnergy *this.maxHerbivoreEnergy * 1.7)){
			double newX,newY;
			newX = Math.max(this.currentLocation.getX() + this.maxSpeed*(2*rand.nextDouble()-1),0);
			newY = Math.max(this.currentLocation.getY() + this.maxSpeed*(2*rand.nextDouble()-1),0);
			newX = Math.min(newX, ekosystemWidth);
			newY = Math.min(newY, ekosystemHeight);
			this.currentLocation.setCoordinates(newX, newY);
		}else if(closestPlant!=null){
			consumePlant(closestPlant,ekosystemWidth,ekosystemHeight, rand);
		}else{
			return null;
		}
		if(rand.nextDouble()<this.chanceOfReproduction && this.currentEnergy>energyForReproduction*1.2){
			LinkedList<BasicHerbivore> newOffsprings = new LinkedList<BasicHerbivore>();
			BasicHerbivore offspring = new BasicHerbivore(this.maxSpeed,this.maxHerbivoreEnergy,startingEnergyOffsprings,this.chanceOfReproduction, this.agePerIteration,this.maxAge,this.energyForReproduction,new Location(this.currentLocation.getX(),this.currentLocation.getY()));
			newOffsprings.add(offspring);
			this.currentEnergy-=this.energyForReproduction;
			return newOffsprings;
		}
		return null;
	}

	/**
	 * Method is used to move towards and consume the target plant if possible
	 * @param closestPlant plant that the herbivore is trying to consume
	 * @param ecosystemWidth width of ecosystem
	 * @param ecosystemHeight height of ecosystem
	 * @param rand random number generator
	 */
	private void consumePlant(AbstractPlant<AbstractPlant> closestPlant,int ecosystemWidth, int ecosystemHeight, Random rand) {
		double plantX=closestPlant.getLocation().getX(), plantY = closestPlant.getLocation().getY();
		double herbX = this.currentLocation.getX(),herbY = this.currentLocation.getY();
		double newX,newY;
		double xDiff = herbX-plantX, yDiff = herbY-plantY;
		double distance = Math.sqrt(xDiff*xDiff+yDiff*yDiff);
		if(distance<5){
			this.currentEnergy+=closestPlant.consume(Math.min(20,(this.maxHerbivoreEnergy-this.currentEnergy)));
			return;
		}
		double directionX = herbX>plantX ? -1 : 1;
		double directionY = herbY>plantY ? -1 : 1;
		double koeff = Math.abs(yDiff/xDiff);
		double movement = this.maxSpeed*rand.nextDouble();
		double radians = Math.atan(koeff);
		newX = herbX+Math.cos(radians)*movement*directionX;
		newY = herbY+Math.sin(radians)*movement*directionY;
		newX = Math.max(newX, 0);
		newY = Math.max(newY, 0);
		this.currentLocation.setCoordinates(Math.min(newX, ecosystemWidth), Math.min(newY, ecosystemHeight));

	}

	@Override
	public boolean isDead() {
		return this.currentAge>this.maxAge || this.currentEnergy<=0;
	}

}
