package artificialLifeSimulator.lifeForms.populations;

import java.util.LinkedList;
import java.util.Random;

import artificialLifeSimulator.abstractClasses.AbstractHerbivore;
import artificialLifeSimulator.abstractClasses.AbstractHerbivorePopulation;
import artificialLifeSimulator.abstractClasses.AbstractPlant;
import artificialLifeSimulator.abstractClasses.AbstractPlantPopulation;
import artificialLifeSimulator.helperClasses.Location;
import artificialLifeSimulator.lifeForms.basic.BasicHerbivore;
import artificialLifeSimulator.lifeForms.basic.BasicPlant;

/**
 * @author Toni
 * Class implements methods that allow the simulation of a basic herbivore population
 */
public class BasicHerbivorePopulation extends AbstractHerbivorePopulation<BasicHerbivore> {
	private int maxPopulationSize;
	private int currentPopulationSize;
	
	
	/**
	 * Default constructor
	 */
	private BasicHerbivorePopulation(){
		
	}
	
	
	/**
	 * Constructor for basic herbivore population
	 * @param plants plant population in the ecosystem
	 * @param maxSize max population size
	 * @param initialSize initial population size
	 * @param maxSpeed max speed of organisms
	 * @param maxEnergy max energy of organisms
	 * @param startingEnergy starting energy of organisms
	 * @param chanceOfReproduction reproduction chance of organisms
	 * @param agePerIteration age per iteration of organisms
	 * @param maxAge max age of organisms
	 * @param energyForReproduction energy required for reproduction
	 * @param ecosystemWidth ecosystem width
	 * @param ecosystemHeight ecosystem height
	 */
	public BasicHerbivorePopulation(AbstractPlantPopulation<AbstractPlant> plants, int maxSize, int initialSize, double maxSpeed,
			double maxEnergy, double startingEnergy, double chanceOfReproduction, double agePerIteration, double maxAge, double energyForReproduction, int ecosystemWidth,
			int ecosystemHeight) {
		Random rand = new Random();
		this.plants = plants;
		this.population = new LinkedList<>();
		this.maxPopulationSize=maxSize;
		this.currentPopulationSize=initialSize;
		for(int i=0;i<currentPopulationSize;i++){
			int x=rand.nextInt(ecosystemWidth), y=rand.nextInt(ecosystemHeight);
			population.add(new BasicHerbivore(maxSpeed,maxEnergy, startingEnergy, chanceOfReproduction, agePerIteration, maxAge,energyForReproduction,new Location(x, y)));
		}
	}

	@Override
	public void nextIteration(int ecosystemWidth, int ecosystemHeight,Random rand) {
		LinkedList<BasicHerbivore> newHerbivores = new LinkedList<BasicHerbivore>();
		LinkedList<BasicHerbivore> toRemove = new LinkedList<>();
		for(BasicHerbivore temp : this.population){
			if (temp.isDead()){
				toRemove.add(temp);
				this.currentPopulationSize--;
				continue;
			}
			LinkedList<BasicHerbivore> tempOffsprings;
			tempOffsprings=temp.nextIteration(ecosystemWidth, ecosystemHeight, rand, findClosestPlant(temp));
			
			if(tempOffsprings != null && this.currentPopulationSize!=this.maxPopulationSize){
				if(this.currentPopulationSize + newHerbivores.size() <this.maxPopulationSize){
					newHerbivores.addAll(tempOffsprings);
					this.currentPopulationSize+=tempOffsprings.size();
				}else{
					int populationSlotsLeft = this.maxPopulationSize-this.currentPopulationSize;
					for(int j = 0;j<populationSlotsLeft && j<tempOffsprings.size();j++){
						newHerbivores.add(tempOffsprings.get(j));
						this.currentPopulationSize++;
					}
				}
			}
		}
		for(BasicHerbivore temp : toRemove){
			this.population.remove(temp);
		}
		this.population.addAll(newHerbivores);
	}


	/**
	 * Method returns the closest plant to the herbivore
	 * @param herb herbivore used in search
	 * @return the closest plant to the given herbivore
	 */
	private AbstractPlant<AbstractPlant> findClosestPlant(AbstractHerbivore<BasicHerbivore> herb) {
		Location loc = herb.getLocation();
		double minDistance=Double.MAX_VALUE;
		AbstractPlant<AbstractPlant> candidate = null;
		for(AbstractPlant plant : plants.getPopulation()){
			Location plantLoc=plant.getLocation();
			double xDiff = Math.abs(plantLoc.getX()-loc.getX());
			double yDiff = Math.abs(plantLoc.getY()-loc.getY());
			double distance = Math.sqrt(xDiff*xDiff + yDiff*yDiff);
			if(distance<minDistance){
				candidate=plant;
				minDistance=distance;
			}
		}
		return candidate;
	}


	@Override
	public AbstractHerbivorePopulation copyOf() { 
		BasicHerbivorePopulation population = new BasicHerbivorePopulation();
		population.maxPopulationSize=this.maxPopulationSize;
		population.currentPopulationSize=this.currentPopulationSize;
		population.population = new LinkedList<BasicHerbivore>();
		for(BasicHerbivore temp : this.population){
			population.population.add((BasicHerbivore)temp.copyOf());
		}
		return population;
	}



}
