package artificialLifeSimulator.lifeForms.populations;

import java.io.Serializable;
import java.util.Random;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

import artificialLifeSimulator.abstractClasses.AbstractCarnivorePopulation;
import artificialLifeSimulator.abstractClasses.AbstractHabitat;
import artificialLifeSimulator.abstractClasses.AbstractHerbivorePopulation;
import artificialLifeSimulator.abstractClasses.AbstractPlantPopulation;
import artificialLifeSimulator.habitats.BasicHabitat;
import artificialLifeSimulator.helperClasses.AlgorithmConstants;
import javafx.scene.canvas.GraphicsContext;
import javafx.scene.paint.Color;
import javafx.scene.paint.Paint;

/**
 * @author Toni
 * Class implements methods required for simulation of the biome where plants,herbivores and carnivores exist
 */
public class Biome implements Runnable, Serializable{
	private transient BlockingQueue<Biome> drawingQueue;
	private AbstractHabitat domain;
	private AbstractHerbivorePopulation animalHerbivorePopulation;
	private AbstractCarnivorePopulation animalCarnivorePopulation;
	private AbstractPlantPopulation plantPopulation;
	private int ecosystemWidth;
	private int ecosystemHeight;
	private transient volatile Thread exit;
	private volatile boolean block = false;
	
	/**
	 * Constructor for Biome class
	 * @param domain habitat where all organisms live
	 * @param carnivore carnivore population
	 * @param herbivore herbivore population
	 * @param plants plant population
	 * @param ekosystemWidth ecosystem width
	 * @param ekosystemHeight ecosystem height
	 */ 
	public Biome(AbstractHabitat domain, AbstractCarnivorePopulation carnivore, AbstractHerbivorePopulation herbivore, 
			AbstractPlantPopulation plants, int ekosystemWidth, int ekosystemHeight){
		this.domain = domain;
		this.animalCarnivorePopulation = carnivore;
		this.animalHerbivorePopulation = herbivore;
		this.plantPopulation = plants;
		this.ecosystemHeight = ekosystemHeight;
		this.ecosystemWidth = ekosystemWidth;
	}
	
	/**
	 *  Default constructor
	 */
	private Biome(){
		
	}
	/**
	 * Method is used to get the habitat
	 * @return Habitat class
	 */
	public AbstractHabitat getDomain(){
		return domain;
	}
	
	/**
	 * Method is used to get the plant population
	 * @return the plant population
	 */
	public AbstractPlantPopulation getPlants(){
		return plantPopulation;
	}
	
	/**
	 * Method is used to get the herbivore population
	 * @return the herbivore population
	 */
	public AbstractHerbivorePopulation getHerbivores(){
		return animalHerbivorePopulation;
	}
	
	/**
	 * Method is used to get the carnivore population
	 * @return the carnivore population
	 */
	public AbstractCarnivorePopulation getCarnivores(){
		return animalCarnivorePopulation;
	}
	
	
	/**
	 * Method is used to draw the biome in three layers
	 * @param topLayer_g graphic context is used to draw the carnivores and herbivores
	 * @param midLayer_g graphic context is used to draw the plants
	 * @param bottomLayer_g graphic context is used to draw the habitat
	 */
	public void drawBiome(GraphicsContext topLayer_g,GraphicsContext midLayer_g,GraphicsContext bottomLayer_g){
		
		if(domain!= null){
			domain.drawHabitat(bottomLayer_g);
		}
		if(plantPopulation !=null){
			plantPopulation.drawPopulation(midLayer_g);
		}
		if(animalHerbivorePopulation != null){
			animalHerbivorePopulation.drawPopulation(topLayer_g);
		}
		if(animalCarnivorePopulation != null){
			animalCarnivorePopulation.drawPopulation(topLayer_g);
		}
	}
	
	/**
	 * Method is used to get ecosytem width
	 * @return ecosystem width
	 */
	public int getEcosystemWidth(){
		return ecosystemWidth;
	}

	/**
	 * Method is used to get ecosytem height
	 * @return ecosystem height
	 */
	public int getEcosystemHeight(){
		return ecosystemHeight;
	}
	
	/**
	 * Method is used to terminate the simulation
	 */
	public void stop(){
		exit = null;
	}
	
	/**
	 *  Method is used to pause the simulation
	 */
	public void block(){
		block=true;
	}
	
	/**
	 * Method is used to set the blocking queue used in communication between the drawing thread and the simulation
	 * @param drawingQueue blocking queue used for communication
	 */
	public void setBlockingQueue(BlockingQueue<Biome> drawingQueue){
		this.drawingQueue = drawingQueue;
	}

	/**
	 * Method is used to return the full copy of the biome
	 * @return copy of the biome
	 */
	public Biome copyOf() {
		Biome copyBiome = new Biome();
		copyBiome.animalCarnivorePopulation = animalCarnivorePopulation == null ? null : this.animalCarnivorePopulation.copyOf();
		copyBiome.animalHerbivorePopulation = animalHerbivorePopulation == null ? null : this.animalHerbivorePopulation.copyOf();
		copyBiome.plantPopulation = plantPopulation == null ? null :this.plantPopulation.copyOf();
		copyBiome.domain = domain == null ? null :this.domain.copyOf();
		copyBiome.block=false;
		copyBiome.ecosystemHeight=this.ecosystemHeight;
		copyBiome.ecosystemWidth=this.ecosystemWidth;
		return copyBiome;
	}
	
	
	/**
	 * Method is used to return the copy of the biome required for proper drawing
	 * @return copy of the biome with information required for drawing
	 */
	public Biome copyForDrawing(){
		Biome copyBiome = new Biome();
		copyBiome.animalCarnivorePopulation = animalCarnivorePopulation == null ? null : this.animalCarnivorePopulation.copyOf();
		copyBiome.animalHerbivorePopulation = animalHerbivorePopulation == null ? null : this.animalHerbivorePopulation.copyOf();
		copyBiome.plantPopulation = plantPopulation == null ? null :this.plantPopulation.copyOf();
		copyBiome.domain = domain == null ? null :this.domain.copyOf();
		copyBiome.block = false;
		copyBiome.ecosystemHeight=this.ecosystemHeight;
		copyBiome.ecosystemWidth=this.ecosystemWidth;
		return copyBiome;
	}

	/**
	 * Method is used to simulate the ecosystem with only plants
	 * @param rand random number generator
	 */
	private void plantsOnly(Random rand){
		exit = Thread.currentThread();
		Thread firstThread = Thread.currentThread();
		while(exit == firstThread){
			plantPopulation.nextIteration(ecosystemWidth, ecosystemHeight, rand);
			Biome temp = copyForDrawing();
			if(block && exit!=null){
				try {
					Thread.currentThread().sleep(Long.MAX_VALUE);
				} catch (InterruptedException e) {
					block = false;
				}
			}
			
			try {
				drawingQueue.offer(temp, AlgorithmConstants.QueueWaitTime, TimeUnit.SECONDS);
			} catch (InterruptedException e) {
				block = false;
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}

	/**
	 * Method is used to simulate the ecosystem with plants and herbivores
	 * @param rand random number generator
	 */
	private void plantsWithHerbivores(Random rand){
		exit = Thread.currentThread();
		Thread firstThread = Thread.currentThread();
		while(exit == firstThread){
			plantPopulation.nextIteration(this.ecosystemWidth,this.ecosystemHeight, rand);
			animalHerbivorePopulation.nextIteration(this.ecosystemWidth,this.ecosystemHeight, rand);
			Biome temp = copyForDrawing();

			if(block && exit!=null){
				try {
					Thread.currentThread().sleep(Long.MAX_VALUE);
				} catch (InterruptedException e) {
					block = false;
				}
			}


			try {
				drawingQueue.offer(temp, AlgorithmConstants.QueueWaitTime, TimeUnit.SECONDS);
			} catch (InterruptedException e) {
				block = false;
			}
		}
	}

	/**
	 * Method is used to simulate the ecosystem with all supported species
	 * @param rand random number generator
	 */
	private void everySpecies(Random rand){
		exit = Thread.currentThread();
		Thread firstThread = Thread.currentThread();
		while(exit == firstThread){
			plantPopulation.nextIteration(ecosystemWidth, ecosystemHeight, rand);
			animalHerbivorePopulation.nextIteration(ecosystemWidth, ecosystemHeight, rand);
			animalCarnivorePopulation.nextIteration(ecosystemWidth, ecosystemHeight, rand);
			Biome temp = copyForDrawing();

			if(block && exit!=null){
				try {
					Thread.currentThread().sleep(Long.MAX_VALUE);
				} catch (InterruptedException e) {
					block = false;
				}
			}
			
			try {
				drawingQueue.offer(temp, AlgorithmConstants.QueueWaitTime, TimeUnit.SECONDS);
			} catch (InterruptedException e) {
				block = false;
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
		}
	}
	
	
	@Override
	public void run() {
		Random rand = new Random();
		if(plantPopulation !=null && animalHerbivorePopulation == null && animalCarnivorePopulation == null){
			plantsOnly(rand);
		}
		if(plantPopulation !=null && animalHerbivorePopulation != null && animalCarnivorePopulation == null){
			plantsWithHerbivores(rand);
		}
		if(plantPopulation !=null && animalHerbivorePopulation!= null && animalCarnivorePopulation != null){
			everySpecies(rand);
		}
	}
	@Override
	public String toString(){
		return "Plant population size: " + this.plantPopulation.getPopulation().size() + 
				"\nHerbivore population size: " + this.animalHerbivorePopulation.getPopulation().size();
	}

	/**
	 * Method is used to set the references between populations
	 */
	public void setReferences() {
		if(this.animalHerbivorePopulation!=null){
			this.animalHerbivorePopulation.setPlants(this.plantPopulation);
			if(this.animalCarnivorePopulation!=null){
				this.animalHerbivorePopulation.setCarnivores(this.animalCarnivorePopulation);
			}
		}
		
	}
}
